// Copyright (c) Jupyter Development Team.
// Distributed under the terms of the Modified BSD License.

import { Token } from '@lumino/coreutils';

/**
 * The path tracker token.
 */
export const IMiddleToken = new Token('@jupyterlab/example-federated-middle');
