/*
 * Copyright (c) Jupyter Development Team.
 * Distributed under the terms of the Modified BSD License.
 */

// We use our own declaration because the existing typings do not work with webpack 5

declare module 'mini-css-extract-plugin';
