/*-----------------------------------------------------------------------------
| Copyright (c) Jupyter Development Team.
| Distributed under the terms of the Modified BSD License.
|----------------------------------------------------------------------------*/

import '@jupyterlab/application/style/scrollbar.css';
import '@jupyterlab/apputils/style/index.js';
import '@jupyterlab/cells/style/index.js';
import '@jupyterlab/codemirror/style/index.js';
import '@jupyterlab/notebook/style/index.js';
import '@jupyterlab/outputarea/style/index.js';
import '@jupyterlab/rendermime/style/index.js';
