/* Syntax highlighting styles for CodeMirror editors in the merge view */

/* Python */
.cm-s-text-x-python .cm-keyword {
    color: #0000ff;
}

.cm-s-text-x-python .cm-comment {
    color: #008000;
}

.cm-s-text-x-python .cm-string {
    color: #a31515;
}

.cm-s-text-x-python .cm-number {
    color: #098658;
}

.cm-s-text-x-python .cm-operator {
    color: #000000;
}

.cm-s-text-x-python .cm-def {
    color: #795e26;
}

.cm-s-text-x-python .cm-variable {
    color: #001080;
}

.cm-s-text-x-python .cm-property {
    color: #001080;
}

/* JavaScript/TypeScript */
.cm-s-text-javascript .cm-keyword,
.cm-s-text-typescript .cm-keyword {
    color: #0000ff;
}

.cm-s-text-javascript .cm-comment,
.cm-s-text-typescript .cm-comment {
    color: #008000;
}

.cm-s-text-javascript .cm-string,
.cm-s-text-typescript .cm-string {
    color: #a31515;
}

.cm-s-text-javascript .cm-number,
.cm-s-text-typescript .cm-number {
    color: #098658;
}

.cm-s-text-javascript .cm-operator,
.cm-s-text-typescript .cm-operator {
    color: #000000;
}

.cm-s-text-javascript .cm-def,
.cm-s-text-typescript .cm-def {
    color: #795e26;
}

.cm-s-text-javascript .cm-variable,
.cm-s-text-typescript .cm-variable {
    color: #001080;
}

.cm-s-text-javascript .cm-property,
.cm-s-text-typescript .cm-property {
    color: #001080;
}

/* JSON */
.cm-s-application-json .cm-property {
    color: #0451a5;
}

.cm-s-application-json .cm-string {
    color: #a31515;
}

.cm-s-application-json .cm-number {
    color: #098658;
}

.cm-s-application-json .cm-keyword {
    color: #0000ff;
}

/* HTML */
.cm-s-text-html .cm-tag {
    color: #800000;
}

.cm-s-text-html .cm-attribute {
    color: #ff0000;
}

.cm-s-text-html .cm-string {
    color: #0000ff;
}

.cm-s-text-html .cm-comment {
    color: #008000;
}

/* CSS */
.cm-s-text-css .cm-property {
    color: #ff0000;
}

.cm-s-text-css .cm-qualifier {
    color: #800000;
}

.cm-s-text-css .cm-variable-2 {
    color: #0000ff;
}

.cm-s-text-css .cm-number {
    color: #098658;
}

.cm-s-text-css .cm-string {
    color: #a31515;
}

.cm-s-text-css .cm-comment {
    color: #008000;
}

/* Markdown */
.cm-s-text-markdown .cm-header {
    color: #0000ff;
}

.cm-s-text-markdown .cm-quote {
    color: #008000;
}

.cm-s-text-markdown .cm-link {
    color: #0000ff;
    text-decoration: underline;
}

.cm-s-text-markdown .cm-string {
    color: #a31515;
}

.cm-s-text-markdown .cm-comment {
    color: #008000;
}

/* C/C++ */
.cm-s-text-x-c .cm-keyword,
.cm-s-text-x-c-src .cm-keyword {
    color: #0000ff;
}

.cm-s-text-x-c .cm-comment,
.cm-s-text-x-c-src .cm-comment {
    color: #008000;
}

.cm-s-text-x-c .cm-string,
.cm-s-text-x-c-src .cm-string {
    color: #a31515;
}

.cm-s-text-x-c .cm-number,
.cm-s-text-x-c-src .cm-number {
    color: #098658;
}

.cm-s-text-x-c .cm-operator,
.cm-s-text-x-c-src .cm-operator {
    color: #000000;
}

.cm-s-text-x-c .cm-def,
.cm-s-text-x-c-src .cm-def {
    color: #795e26;
}

.cm-s-text-x-c .cm-variable,
.cm-s-text-x-c-src .cm-variable {
    color: #001080;
}

.cm-s-text-x-c .cm-property,
.cm-s-text-x-c-src .cm-property {
    color: #001080;
}

/* Default for plain text */
.cm-s-text-plain .cm-string {
    color: #a31515;
}

.cm-s-text-plain .cm-number {
    color: #098658;
}

.cm-s-text-plain .cm-comment {
    color: #008000;
}

/* Dark theme support */
.jp-mod-dark .cm-s-text-x-python .cm-keyword,
.jp-mod-dark .cm-s-text-javascript .cm-keyword,
.jp-mod-dark .cm-s-text-typescript .cm-keyword {
    color: #569cd6;
}

.jp-mod-dark .cm-s-text-x-python .cm-comment,
.jp-mod-dark .cm-s-text-javascript .cm-comment,
.jp-mod-dark .cm-s-text-typescript .cm-comment {
    color: #6a9955;
}

.jp-mod-dark .cm-s-text-x-python .cm-string,
.jp-mod-dark .cm-s-text-javascript .cm-string,
.jp-mod-dark .cm-s-text-typescript .cm-string {
    color: #ce9178;
}

.jp-mod-dark .cm-s-text-x-python .cm-number,
.jp-mod-dark .cm-s-text-javascript .cm-number,
.jp-mod-dark .cm-s-text-typescript .cm-number {
    color: #b5cea8;
}

.jp-mod-dark .cm-s-text-x-python .cm-def,
.jp-mod-dark .cm-s-text-javascript .cm-def,
.jp-mod-dark .cm-s-text-typescript .cm-def {
    color: #dcdcaa;
}

.jp-mod-dark .cm-s-text-x-python .cm-variable,
.jp-mod-dark .cm-s-text-javascript .cm-variable,
.jp-mod-dark .cm-s-text-typescript .cm-variable {
    color: #9cdcfe;
}

.jp-mod-dark .cm-s-text-x-python .cm-property,
.jp-mod-dark .cm-s-text-javascript .cm-property,
.jp-mod-dark .cm-s-text-typescript .cm-property {
    color: #9cdcfe;
}

.jp-mod-dark .cm-s-application-json .cm-property {
    color: #9cdcfe;
}

.jp-mod-dark .cm-s-application-json .cm-string {
    color: #ce9178;
}

.jp-mod-dark .cm-s-application-json .cm-number {
    color: #b5cea8;
}