/*-----------------------------------------------------------------------------
| Copyright (c) Jupyter Development Team.
| Distributed under the terms of the Modified BSD License.
|----------------------------------------------------------------------------*/

@import './commandpalette.css';
@import './dialog.css';
@import './inputdialog.css';
@import './licenses.css';
@import './mainareawidget.css';
@import './materialcolors.css';
@import './toolbar.css';
