// Copyright (c) Jupyter Development Team.
// Distributed under the terms of the Modified BSD License.

export {
  /**
   * @deprecated Debugger icons are now in the ui-components package. Import from '@jupyterlab/ui-components' instead.
   * Use: import { closeAllIcon } from '@jupyterlab/ui-components';
   */
  closeAllIcon,
  /**
   * @deprecated Debugger icons are now in the ui-components package. Import from '@jupyterlab/ui-components' instead.
   * Use: import { runIcon } from '@jupyterlab/ui-components';
   */
  runIcon as continueIcon,
  /**
   * @deprecated Debugger icons are now in the ui-components package. Import from '@jupyterlab/ui-components' instead.
   * Use: import { exceptionsIcon } from '@jupyterlab/ui-components';
   */
  exceptionsIcon as exceptionIcon,
  /**
   * @deprecated Debugger icons are now in the ui-components package. Import from '@jupyterlab/ui-components' instead.
   * Use: import { openKernelSourceIcon } from '@jupyterlab/ui-components';
   */
  openKernelSourceIcon,
  /**
   * @deprecated Debugger icons are now in the ui-components package. Import from '@jupyterlab/ui-components' instead.
   * Use: import { pauseIcon } from '@jupyterlab/ui-components';
   */
  pauseIcon,
  /**
   * @deprecated Debugger icons are now in the ui-components package. Import from '@jupyterlab/ui-components' instead.
   * Use: import { stepIntoIcon } from '@jupyterlab/ui-components';
   */
  stepIntoIcon,
  /**
   * @deprecated Debugger icons are now in the ui-components package. Import from '@jupyterlab/ui-components' instead.
   * Use: import { stepOutIcon } from '@jupyterlab/ui-components';
   */
  stepOutIcon,
  /**
   * @deprecated Debugger icons are now in the ui-components package. Import from '@jupyterlab/ui-components' instead.
   * Use: import { stepOverIcon } from '@jupyterlab/ui-components';
   */
  stepOverIcon,
  /**
   * @deprecated Debugger icons are now in the ui-components package. Import from '@jupyterlab/ui-components' instead.
   * Use: import { stopIcon } from '@jupyterlab/ui-components';
   */
  stopIcon as terminateIcon,
  /**
   * @deprecated Debugger icons are now in the ui-components package. Import from '@jupyterlab/ui-components' instead.
   * Use: import { variableIcon } from '@jupyterlab/ui-components';
   */
  variableIcon,
  /**
   * @deprecated Debugger icons are now in the ui-components package. Import from '@jupyterlab/ui-components' instead.
   * Use: import { viewBreakpointIcon } from '@jupyterlab/ui-components';
   */
  viewBreakpointIcon
} from '@jupyterlab/ui-components';
